
/* Chrysalide - Outil d'analyse de fichiers binaires
 * coproc.c - décallages de valeurs
 *
 * Copyright (C) 2016 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "coproc.h"


#include "../../../operand-int.h"



/* Définition d'un opérande représentant un co-processeur (instance) */
struct _GArmV7CoprocOperand
{
    GArchOperand parent;                    /* Instance parente            */

    uint8_t index;                          /* Indice du co-processeur     */

};


/* Définition d'un opérande représentant un co-processeur (classe) */
struct _GArmV7CoprocOperandClass
{
    GArchOperandClass parent;               /* Classe parente              */

};


/* Initialise la classe des coprocs de domaine et d'accès. */
static void g_armv7_coproc_operand_class_init(GArmV7CoprocOperandClass *);

/* Initialise une instance de coproc de domaine et d'accès. */
static void g_armv7_coproc_operand_init(GArmV7CoprocOperand *);

/* Supprime toutes les références externes. */
static void g_armv7_coproc_operand_dispose(GArmV7CoprocOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_armv7_coproc_operand_finalize(GArmV7CoprocOperand *);

/* Traduit un opérande en version humainement lisible. */
static void g_armv7_coproc_operand_print(const GArmV7CoprocOperand *, GBufferLine *, AsmSyntax);



/* Indique le type défini par la GLib pour un co-processeur ARM. */
G_DEFINE_TYPE(GArmV7CoprocOperand, g_armv7_coproc_operand, G_TYPE_ARCH_OPERAND);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des co-processeurs ARM.                 *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_coproc_operand_class_init(GArmV7CoprocOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchOperandClass *operand;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);
    operand = G_ARCH_OPERAND_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_armv7_coproc_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_armv7_coproc_operand_finalize;

    operand->print = (operand_print_fc)g_armv7_coproc_operand_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance de co-processeur ARM.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_coproc_operand_init(GArmV7CoprocOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_coproc_operand_dispose(GArmV7CoprocOperand *operand)
{
    G_OBJECT_CLASS(g_armv7_coproc_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_coproc_operand_finalize(GArmV7CoprocOperand *operand)
{
    G_OBJECT_CLASS(g_armv7_coproc_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : raw = valeur brute du co-processeur à considérer.            *
*                                                                             *
*  Description : Crée une représentation d'un co-processeur ARM.              *
*                                                                             *
*  Retour      : Opérande mis en place.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_coproc_operand_new(uint8_t raw)
{
    GArmV7CoprocOperand *result;            /* Structure à retourner       */

    result = g_object_new(G_TYPE_ARMV7_COPROC_OPERAND, NULL);

    result->index = raw;

    return G_ARCH_OPERAND(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_coproc_operand_print(const GArmV7CoprocOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    char name[5];                           /* Mot clef principal          */
    size_t nlen;                            /* Taille de ce mot clef       */

    nlen = snprintf(name, sizeof(name), "p%hhu", operand->index);

    g_buffer_line_append_text(line, BLC_ASSEMBLY, name, nlen, RTT_REGISTER, NULL);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Fournit l'indice d'un co-processeur ARM.                     *
*                                                                             *
*  Retour      : Inditifiant représentant le co-processeur.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

uint8_t g_armv7_coproc_operand_get_index(const GArmV7CoprocOperand *operand)
{
    return operand->index;

}
