
/* Chrysalide - Outil d'analyse de fichiers binaires
 * arithm.c - décompilation des opérations arithmétiques
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "translate.h"


#include "../instruction.h"
#include "../../../decomp/expr/arithm.h"
#include "../../../decomp/expr/assign.h"
#include "../../../decomp/expr/immediate.h"



/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction d'origine à convertir.                   *
*                ctx   = contexte de la phase de décompilation.               *
*                                                                             *
*  Description : Décompile une instruction de type 'opérations arithmétiques'.*
*                                                                             *
*  Retour      : Instruction mise en place ou NULL.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *dalvik_decomp_instr_arithm(const GArchInstruction *instr, GDecContext *ctx)
{
    GDecInstruction *result;                /* Instruction à retourner     */
    ArithmOperationType type;               /* Type d'opération menée      */
    vmpa_t addr;                            /* Adresse de l'instruction    */
    GArchOperand *operand;                  /* Opérande de l'instruction   */
    GDecInstruction *op1;                   /* Premier opérande utilisé    */
    GDecInstruction *op2;                   /* Second opérande utilisé     */
    GDecInstruction *dest;                  /* Enregistrement du résultat  */
    GDecInstruction *arithm;                /* Opération arithmétique      */

    switch (g_dalvik_instruction_get_opcode(G_DALVIK_INSTRUCTION(instr)))
    {
        case DOP_ADD_INT:
            type = AOT_ADD;
            break;
        case DOP_SUB_INT:
            type = AOT_SUB;
            break;
        case DOP_MUL_INT:
            type = AOT_MUL;
            break;
        case DOP_DIV_INT:
            type = AOT_DIV;
            break;
        case DOP_REM_INT:
            type = AOT_REM;
            break;
        case DOP_AND_INT:
            type = AOT_AND;
            break;
        case DOP_OR_INT:
            type = AOT_OR;
            break;
        case DOP_XOR_INT:
            type = AOT_XOR;
            break;
        default:
            type = AOT_COUNT;
            break;
    }

    g_arch_instruction_get_location(instr, NULL, NULL, &addr);

    operand = g_arch_instruction_get_operand(instr, 1);
    op1 = g_dec_context_convert_register(ctx, operand, false, addr);

    operand = g_arch_instruction_get_operand(instr, 2);
    op2 = g_dec_context_convert_register(ctx, operand, false, addr);

    operand = g_arch_instruction_get_operand(instr, 0);
    dest = g_dec_context_convert_register(ctx, operand, true, addr);

    arithm = g_arithm_expression_new(G_DEC_EXPRESSION(op1), type, G_DEC_EXPRESSION(op2));
    result = g_assign_expression_new(G_DEC_EXPRESSION(dest), G_DEC_EXPRESSION(arithm));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction d'origine à convertir.                   *
*                ctx   = contexte de la phase de décompilation.               *
*                                                                             *
*  Description : Décompile une instruction de type 'opérations arithmétiques'.*
*                                                                             *
*  Retour      : Instruction mise en place ou NULL.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *dalvik_decomp_instr_arithm_2addr(const GArchInstruction *instr, GDecContext *ctx)
{
    GDecInstruction *result;                /* Instruction à retourner     */
    ArithmOperationType type;               /* Type d'opération menée      */
    vmpa_t addr;                            /* Adresse de l'instruction    */
    GArchOperand *operand;                  /* Opérande de l'instruction   */
    GDecInstruction *op1;                   /* Premier opérande utilisé    */
    GDecInstruction *op2;                   /* Second opérande utilisé     */
    GDecInstruction *dest;                  /* Enregistrement du résultat  */
    GDecInstruction *arithm;                /* Opération arithmétique      */

    switch (g_dalvik_instruction_get_opcode(G_DALVIK_INSTRUCTION(instr)))
    {
        case DOP_ADD_INT_2ADDR:
            type = AOT_ADD;
            break;
        case DOP_MUL_INT_2ADDR:
        case DOP_MUL_DOUBLE_2ADDR:
            type = AOT_MUL;
            break;
        case DOP_DIV_INT_2ADDR:
            type = AOT_DIV;
            break;
        case DOP_REM_INT_2ADDR:
            type = AOT_REM;
            break;
        case DOP_AND_INT_2ADDR:
            type = AOT_AND;
            break;
        case DOP_OR_INT_2ADDR:
            type = AOT_OR;
            break;
        case DOP_XOR_INT_2ADDR:
            type = AOT_XOR;
            break;
        default:
            type = AOT_COUNT;
            break;
    }

    g_arch_instruction_get_location(instr, NULL, NULL, &addr);

    operand = g_arch_instruction_get_operand(instr, 1);
    op2 = g_dec_context_convert_register(ctx, operand, false, addr);

    operand = g_arch_instruction_get_operand(instr, 0);
    op1 = g_dec_context_convert_register(ctx, operand, false, addr);
    dest = g_dec_context_convert_register(ctx, operand, true, addr);

    arithm = g_arithm_expression_new(G_DEC_EXPRESSION(op1), type, G_DEC_EXPRESSION(op2));
    result = g_assign_expression_new(G_DEC_EXPRESSION(dest), G_DEC_EXPRESSION(arithm));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction d'origine à convertir.                   *
*                ctx   = contexte de la phase de décompilation.               *
*                                                                             *
*  Description : Décompile une instruction de type 'opérations arithmétiques'.*
*                                                                             *
*  Retour      : Instruction mise en place ou NULL.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *dalvik_decomp_instr_arithm_lit(const GArchInstruction *instr, GDecContext *ctx)
{
    GDecInstruction *result;                /* Instruction à retourner     */
    ArithmOperationType type;               /* Type d'opération menée      */
    vmpa_t addr;                            /* Adresse de l'instruction    */
    GArchOperand *operand;                  /* Opérande de l'instruction   */
    GDecInstruction *op1;                   /* Premier opérande utilisé    */
    GDecInstruction *op2;                   /* Second opérande utilisé     */
    GDecInstruction *dest;                  /* Enregistrement du résultat  */
    GDecInstruction *arithm;                /* Opération arithmétique      */

    switch (g_dalvik_instruction_get_opcode(G_DALVIK_INSTRUCTION(instr)))
    {
        case DOP_ADD_INT_LIT8:
        case DOP_ADD_INT_LIT16:
            type = AOT_ADD;
            break;
        case DOP_MUL_INT_LIT8:
        case DOP_MUL_INT_LIT16:
            type = AOT_MUL;
            break;
        case DOP_DIV_INT_LIT8:
        case DOP_DIV_INT_LIT16:
            type = AOT_DIV;
            break;
        case DOP_REM_INT_LIT8:
        case DOP_REM_INT_LIT16:
            type = AOT_REM;
            break;
        case DOP_AND_INT_LIT8:
        case DOP_AND_INT_LIT16:
            type = AOT_AND;
            break;
        case DOP_OR_INT_LIT8:
        case DOP_OR_INT_LIT16:
            type = AOT_OR;
            break;
        case DOP_XOR_INT_LIT8:
        case DOP_XOR_INT_LIT16:
            type = AOT_XOR;
            break;
        default:
            type = AOT_COUNT;
            break;
    }

    g_arch_instruction_get_location(instr, NULL, NULL, &addr);

    operand = g_arch_instruction_get_operand(instr, 1);
    op1 = g_dec_context_convert_register(ctx, operand, false, addr);

    operand = g_arch_instruction_get_operand(instr, 2);
    op2 = g_imm_expression_new(G_IMM_OPERAND(operand));

    operand = g_arch_instruction_get_operand(instr, 0);
    dest = g_dec_context_convert_register(ctx, operand, true, addr);

    arithm = g_arithm_expression_new(G_DEC_EXPRESSION(op1), type, G_DEC_EXPRESSION(op2));
    result = g_assign_expression_new(G_DEC_EXPRESSION(dest), G_DEC_EXPRESSION(arithm));

    return result;

}
